--------------------NOTICE-------------------------------
-- Software ID: tk_mem.e
-- Version:     0.56
-- Copyright:   (c) 2000 David Cuny, Derek Parnell
--              All rights reserved.
-- Licence:            
-- This software is provided 'as-is', without any express or implied warranty.
-- In no event will the authors be held liable for any damages arising from
-- the use of this software.
--
-- Permission is granted to anyone to use this software for any purpose,
-- including commercial applications, and to alter it and redistribute it
-- freely, subject to the following restrictictions:
-- 1. The origin of this software must not be misrepresented; you must not
--    claim that you wrote the original software.
-- 2. If you use this software in a product, acknowedgement in the product's
--    documenation and binary are required.
-- 3. Altered source versions, and works substantially derived from the it,
--    must...
--   a) be plainly be marked as such,
--   b) not be misrepresented as the original software,
--   c) include this notice, unaltered.
--------------------End of NOTICE------------------------
-- Last Changed on 4/06/2001 at 21:52:34 by Derek Parnell --

-- The library was originally developed by David Cuny, and over time, a 
-- number of other people have made additions and corrections. Where
-- appropriate the work that other people have contributed is marked thus...

--     Begin <person's name>
--     End <person's name>

-----------------------------------------------------------------------------
-- Generic Tools
-- These are routines that I use often enough to wish they were part of
-- the standard Euphoria library

--/topic Memory Management routines
--/info
--Low-Level Memory management routines
without warning
without trace
without profile
include machine.e
include dll.e
include tk_misc.e

constant
kernel32                      = open_dll( "kernel32.dll" ),
xIsBadWritePtr = define_c_func(kernel32, "IsBadWritePtr", {C_LONG, C_LONG}, C_LONG),
xIsBadReadPtr = define_c_func(kernel32, "IsBadReadPtr", {C_LONG, C_LONG}, C_LONG),
xIsBadCodePtr = define_c_func(kernel32, "IsBadReadPtr", {C_LONG, C_LONG}, C_LONG),
xlstrlen     = define_c_func(kernel32,"lstrlen",{C_POINTER},C_INT)


--CRITICAL_SECTION CriticalSection;

--// Initialize the critical section.
--InitializeCriticalSection(&CriticalSection);

--// Request ownership of the critical section.
--    EnterCriticalSection(&CriticalSection);

    -- Access the shared resource.


    -- Release ownership of the critical section.
    --LeaveCriticalSection(&CriticalSection);

    -- Release resources used by the critical section object.
    --DeleteCriticalSection(&CriticalSection)

-- stores abort handler's routine id
integer  vAbortRtn  vAbortRtn  = -1

-- The number of allocations
integer vAllocations vAllocations = 0    
sequence vExtraCare vExtraCare = {}

-- List of memset ids
sequence vOwners vOwners = {}

-- List of address sets. Each set has a list of addresses.
sequence vSets   vSets = {}

-- stores the accumulated size of a structure as it is being defined.
integer  vAllotted  vAllotted  = 0

-- This is added to calls to alloc.
integer vSafetyBuffer vSafetyBuffer = 4

--/topic Memory Management Routines
--/Var UsingSafeCode
--/desc Determines whether or not the 'safe' versions of peek and poke are used.
--This is primarily as debugging aid. You only need to set this if you suspect
--that your program is causing memory corruptions or accessing strange locations.
--
-- Set this to zero to turn off the safe versions. Any other value causes
-- the functions /SafePeek(), /SafePeek4s(), /SafePeek4u(), /SafePoke(), and /SafePoke4()
-- to perform memory access tests prior to fetching or changing memory.
--
-- The initial setting is 0. That is, the safe versions are not being used.
--
--Example:
--/code
--     --Ensure I can change RAM safely
--     UsingSafeCode = 1
--     SafePoke4( adr, 0)
--/endcode

global integer UsingSafeCode UsingSafeCode = 0

--/topic Memory Management Routines
--/func llSetAbort( i )
--/desc Sets the routine id of an Abort routine.
--/ret The previous value set.
--
-- Used to indicate if an error routine needs to be
-- called in the event of a catastophic error.
-- The error routine is assumed to be a procedure
-- that accepts a single sequence (typically an 
-- message string).
--
-- Example:
--
-- /code        
--      integer RtnID, OldID
--      RtnID = routine_id("abortErr")
--      OldID = llSetAbort(RtnID)
-- /endcode

global function llSetAbort(integer i)
-- Set the abort handler id. The routine must take two parameters,
-- a sequence (msg) and a integer (mode ==> 1=warning, 2=fatal)
    integer lOldRtn

    lOldRtn = vAbortRtn
    
    vAbortRtn = i
    
    return lOldRtn
end function

--/topic Memory Management Routines
--/func llSetSafetyBuffer( integer bufsize )
--/ret Returns the current value of the buffer.
--/desc Sets the number of extra bytes allocated as a safety buffer
--This is primarily a debugging aid. If you suspect that memory overrun
-- problems are happening, you can use this to test the idea.
--
-- /i bufsize must be zero or more. A negative size does not change the
-- current setting. /n
-- The initial setting is 0 bytes. That is, there is no safety buffer.
--
--Example:
--/code
--       -- Cause a 4-byte buffer to be added to all memory allocations.
--       oldval = llSetSafetyBuffer( 4 )
--/endcode
global function llSetSafetyBuffer(integer bufsize)
    integer oldvalue

    oldvalue = vSafetyBuffer
    if bufsize >= 0 then
        vSafetyBuffer = bufsize
    end if

    return oldvalue
    
end function

global function llSetHeapIncrement(integer bufsize)
    -- no longer supported
    return 0
    
end function

procedure myFree(atom pAddress)
    object VOID
    integer lPosn  
    integer lMemSize

    if c_func(xIsBadWritePtr, {pAddress, 1}) then
        return
    end if
    
        
    -- Check if this addr needs extra care checking.
    lPosn = find(pAddress, vExtraCare)
    if lPosn != 0 then 
        pAddress -= 16
        if peek4u(pAddress) != pAddress or
           peek4u(pAddress+4) != pAddress + 4 or
           peek4u(pAddress+8) != #C0D0E0F0 or
           peek4u(pAddress+12) != #8899AABB then
               -- The area prior to the users RAM was corrupted.
                if vAbortRtn >= 0 then
                    call_proc(vAbortRtn, {
                                sprintf("Corrupted RAM at address %d.", pAddress)} )
                else                    
                    crash_message(sprintf("Corrupted RAM at address %d.", pAddress))
                    VOID = 0/0 -- Force a crash
                end if
        end if
        -- Remove it from the list.
        vExtraCare = vExtraCare[1..lPosn-1] & vExtraCare[lPosn+1..length(vExtraCare)]
    end if

    pAddress -= 4
    
    vAllocations -= 1
        
    free(pAddress)
    return

end procedure

function myAllocate(integer pSize)
    atom lAddr
    atom lPreSize       
    integer lMemSize


        
    if UsingSafeCode != 0 then
        lPreSize = 16
    else
        lPreSize = 0
    end if        

    -- Add some bytes for a safety buffer                        
    lMemSize = lPreSize + pSize + 4 + vSafetyBuffer  
    lAddr = allocate(lMemSize)
    
    if lAddr != 0 then
        mem_set( lAddr, 0, lMemSize)
        
        vAllocations += 1

                                   
        -- Save total size for later reuse.
        poke4(lAddr, lMemSize)
        lAddr += 4
        
        -- If being safe, add this address to the extra care list,
        -- so that the free() can see if nothing was messed up.
        if lPreSize != 0 then
            lPreSize = lAddr
            lAddr += 16
            vExtraCare &= lAddr
            poke4(lPreSize, {lPreSize, lPreSize+4, #C0D0E0F0, #8899AABB})
        end if
    end if
        
    return lAddr
end function

--/topic Memory Management Routines
--/func SafePeek(object addr)
--/desc This is an enhanced version of Euphoria's peek() function
--/ret see peek() for details.
--If /i UsingSafeCode is not zero, this does a memory access check first.

global function SafePeek(object addr)
    atom a, l   
    if UsingSafeCode = 0 then
        return peek(addr)
    end if
    
    if sequence(addr) and length(addr) = 2 and atom(addr[1]) and atom(addr[2]) then
        l = addr[2]
        a = addr[1]
    else
        a = addr
        l = 1
    end if

    if c_func(xIsBadReadPtr, {a,l}) = 0 then
        return peek(addr)
    else
        if vAbortRtn >= 0
        then
            call_proc(vAbortRtn, {
                        sprintf("Peek at address %d for length %d failed.",
                                    {a,l})} )
            return 0 -- Just in case the abort does not crash program.
        else
            if object(addr) then
                return {}
            else
                return 0
            end if
        end if
    end if
end function

--/topic Memory Management Routines
--/func SafePeek4s(object addr)
--/desc This is an enhanced version of Euphoria's peek4s() function
--/ret see peek4s() for details.
--If /i UsingSafeCode is not zero, this does a memory access check first.
global function SafePeek4s(object addr)
    atom a, l

    if UsingSafeCode = 0 then
        return peek4s(addr)
    end if
    

    if sequence(addr) and length(addr) = 2 and atom(addr[1]) and atom(addr[2]) then
        l = addr[2] * 4
        a = addr[1]
    else
        a = addr
        l = 4
    end if

    if c_func(xIsBadReadPtr, {a,l}) = 0 then
        return peek4s(addr)
    else
        if vAbortRtn >= 0
        then
            call_proc(vAbortRtn, {
                        sprintf("Peek4s at address %d for count of %d failed.",
                                    {a,l/4})} )
            return 0 -- Just in case the abort does not crash program.
        else
            if object(addr) then
                return {}
            else
                return 0
            end if
        end if
    end if
end function

--/topic Memory Management Routines
--/func SafePeek4u(object addr)
--/desc This is an enhanced version of Euphoria's peek4u() function
--/ret see peek4u() for details.
--If /i UsingSafeCode is not zero, this does a memory access check first.
global function SafePeek4u(object addr)
    atom a, l

    if UsingSafeCode = 0 then
        return peek4u(addr)
    end if
    

    if sequence(addr) and length(addr) = 2 and atom(addr[1]) and atom(addr[2]) then
        l = addr[2] * 4
        a = addr[1]
    else
        a = addr
        l = 4
    end if

    if c_func(xIsBadReadPtr, {a,l}) = 0 then
        return peek4u(addr)
    else
        if vAbortRtn >= 0
        then
            call_proc(vAbortRtn, {
                        sprintf("Peek4u at address %d for count of %d failed.",
                                    {a,l/4})} )
            return 0 -- Just in case the abort does not crash program.
        else
            if object(addr) then
                return {}
            else
                return 0
            end if
        end if
    end if
end function

--/topic Memory Management Routines
--/func SafePoke(object addr)
--/desc This is an enhanced version of Euphoria's poke() function
--/ret see poke() for details.
--If /i UsingSafeCode is not zero, this does a memory access check first.
global procedure SafePoke(atom addr, object thedata)
    atom a, l

    if UsingSafeCode = 0 then
        poke(addr, thedata)
        return
    end if
    

    if sequence(thedata) then
        l = length(thedata)
    else
        l = 1
    end if

    if c_func(xIsBadWritePtr, {addr,l}) = 0 then
        poke(addr, thedata)
    else
        if vAbortRtn >= 0
        then
            call_proc(vAbortRtn, {
                        sprintf("Poke at address %d for length of %d failed.",
                                    {addr,l})} )
            return -- Just in case the abort does not crash program.
        end if
    end if
end procedure

--/topic Memory Management Routines
--/func SafePoke4(object addr)
--/desc This is an enhanced version of Euphoria's poke4() function
--/ret see poke4() for details.
--If /i UsingSafeCode is not zero, this does a memory access check first.
global procedure SafePoke4(atom addr, object thedata)
    atom a, l

    if UsingSafeCode = 0 then
        poke4(addr, thedata)
        return
    end if
    

    if sequence(thedata) then
        l = length(thedata) * 4
    else
        l = 4
    end if

    if c_func(xIsBadWritePtr, {addr,l}) = 0 then
        poke4(addr, thedata)
    else
        if vAbortRtn >= 0
        then
            call_proc(vAbortRtn, {
                        sprintf("Poke4 at address %d for count of %d failed.",
                                    {addr,l/4})} )
            return -- Just in case the abort does not crash program.
        end if
    end if
end procedure

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Memory Management memory management Routines
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------


global constant
    Byte    = -1,  
    Int8    = Byte,
    Word    = -2,
    Integer = Word,
    Int16   = Word,
    Long    = -3,
    DWord   = Long,
    Int32   = Long,
    Lpsz    = -4,
    Hndl    = -5,
    HndlAddr = -6,
    Strz    = -7,
    UInt    = -8,
    Ptr     = UInt,
    ULong   = UInt,
    Single  = -9,
    Float   = Single,
    Double  = -10
    
constant vSizeNames = {Byte, Word, Long, Lpsz, Hndl, HndlAddr, Strz, UInt, Single, Double}
constant vSizeLengs = {   1,    2,    4,    4,    4,        4,    1,    4,      4,      8}

--/topic Memory Management Routines
--/func manage_mem( atom Owner, atom Address)
--/desc Records an acquired memory for garbage collection.
-- Normally this is handled automatically by /acquire_mem() but if you
-- are expected to manage some memory acquired by another means, such as
-- a Windows call or a 'C' routine, you can use this to record the memory for
-- subsequent release by /release_mem().
--
-- Example:
-- /code    
--      atom mset, pt, pstr
--
--      -- Establish a new memory set.
--      mset = /new_memset()
--      -- calls a routine which returns a structure address.
--      pt = c_func( xyz, {abc})
--      -- register this memory
--      manage_mem(mset, pt)
--      . . .
--      give all the memory area in 'mset' back
--      release_mem(mset)
-- /endcode
 
global procedure manage_mem( atom pOwner, atom pAddr )
    -- save location for garbage collection
    -- This also is used to move a block from one memset to another.
    integer lOwnerSub, lAddrSub

    -- Take the address out of any existing memset.
    lAddrSub = 0
    for i = 1 to length(vSets) do
        lAddrSub = find(pAddr, vSets[i])
        if lAddrSub != 0 then
            vSets[i] = removeIndex(lAddrSub, vSets[i])
            exit
        end if
    end for
    
    -- Create a new memset if required.
    lOwnerSub = find(pOwner, vOwners)
    if lOwnerSub = 0 then
        vOwners &= pOwner
        lOwnerSub = length(vOwners)
        vSets = append(vSets, {})
    end if
           
    -- Store the address into the specified memset
    vSets[lOwnerSub] &= pAddr
end procedure

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func acquire_mem( atom Owner, object structure )
--/desc Allocate memory for structure, and initialize to zero.
-- The memory allocated is linked to the /i Owner and all the
-- owner's memory can be released by one call. /n
-- If /i structure is a string, it is copied to the memory
-- location along with a zero byte.
--
-- If /i structure is an atom, it specifies that amount of memory
-- to acquire (a minimum of 4 bytes will always be acquired) and
-- the memory is set to all zeros.
--
--/ret Address of allocated memory.
--
-- Example:
-- /code    
--      atom mset, pt, pstr
--
--      -- Establish a new memory set.
--      mset = /new_memset()
--      -- get enough memory to hold a UInt datatype
--      xy = acquire_mem( UInt )
--      -- allocate a point structure
--      pt = acquire_mem( mset, SIZEOF_POINT )
--      -- copy a Euphoria string to a 'C' string area.
--      pstr = /acquire_mem( mset, "My String Data" )
--      . . .
--      give all the memory area in 'mset' back
--      release_mem(mset)
-- /endcode

global function acquire_mem( atom pOwner, object pData )

        -- allocate space for a structure (mininum of 4 bytes)
        -- and initialize to zero
        
    atom at

    if sequence(pData) then
        -- place string in memory
        at = myAllocate( 1 + length(pData) )
        if at != 0 then
            poke(at, pData)
            poke(at + length(pData) , 0)
        end if
    else
        -- Check for special datatype "names"
        if pData < 0 then
            pData = find(pData, vSizeNames)
            if pData != 0 then
                pData = vSizeLengs[pData]
            end if
        end if

        if pData < 4 then
          pData = 4
        end if

        at  = myAllocate( pData )
        
        
    end if
    
    if at = 0 then
        if vAbortRtn >= 0
        then
            call_proc(vAbortRtn, { "Unable to allocate space."} )
            return 0 -- Just in case the abort does not crash program.
        end if
    else
        manage_mem(pOwner, at)
    end if


    return at
        
end function

--/topic Memory Management Routines
--/proc release_mem( atom structure )
--/desc Returns the memory allocated by /acquire_mem() back to the system.
-- If /i structure is a memory set id, as returned by /new_memset(), then
-- all the memory owned in the memory set is returned and the memory set id
-- is released. That is, it cannot be reused. /n
-- If /i structure is a memory address returned by /acquire_mem(), then just
-- that memory is released. The memory set it belonged to is still usable.
--
-- Example:
-- /code    
--      atom mset, pt, pstr
--
--      -- Establish a new memory set.
--      mset = /new_memset()
--      -- get enough memory to hold a UInt datatype
--      xy = /acquire_mem( UInt )
--      -- allocate a point structure
--      pt = /acquire_mem( mset, SIZEOF_POINT )
--      -- copy a Euphoria string to a 'C' string area.
--      pstr = /acquire_mem( mset, "My String Data" )
--      . . .
--      give all the memory area in 'mset' back
--      release_mem(mset)
-- /endcode

global procedure release_mem( atom pData )
    integer lOwnerSub
    integer lAddrSub, lAddrList
    integer ls, ss, Phase1
    sequence sets

    -- Check for "special" uninitialized memset value.
    if pData = -1 then
        return
    end if

    -- See if this is a memset owner.
    lOwnerSub = find(pData, vOwners)
    
    if lOwnerSub = 0 then
        -- If not a memset, see which memset it belongs to.
        lAddrSub = 0
        for i = 1 to length(vSets) do
            lAddrSub = find(pData, vSets[i])
            if lAddrSub != 0 then
                lAddrList = i
                exit
            end if
        end for

        if lAddrSub = 0 then
            -- Not in any memset!
            if vAbortRtn >= 0 then
                call_proc(vAbortRtn, {"Trying to release unacquired memory"})
                return
            end if
            
        else
            -- Remove it from the memset
            vSets[lAddrList] = removeIndex(lAddrSub,
                                           vSets[lAddrList])
            -- Give the memory back to the system
            myFree(pData)
        end if

        return
    end if

    --
    sets = {pData}
    ss   = 1

    -- Phase 1: Identified the complete hierarchy of owned memory sets.
    Phase1 = 1
    while Phase1 do
        ls = length(sets)
        for i = ss to ls do
            lOwnerSub = find(sets[i], vOwners)
            for j = 1 to length(vSets[lOwnerSub]) do
                lAddrSub = find(vSets[lOwnerSub][j],
                                vOwners)
                if lAddrSub != 0 then
                    sets &= vOwners[lAddrSub]
                end if
            end for
        end for
        ss = ls + 1
        Phase1 = (ls != length(sets))
    end while

    -- Phase 2: Free the memory addresses.
    for i = length(sets) to 1 by -1 do
        lOwnerSub = find(sets[i], vOwners)
        for j = 1 to length(vSets[lOwnerSub]) do
            if vSets[lOwnerSub][j] != 0 then
                myFree(vSets[lOwnerSub][j])

            end if
        end for
    end for
    if sets[1] != 0 then
        myFree(sets[1])

    end if

    -- Phase 3: - clean up the allocation array
    for i = 1 to length(sets) do
        lOwnerSub = find(sets[i], vOwners)
        if lOwnerSub > 0 then
            vSets = removeIndex(lOwnerSub, vSets)
            vOwners = removeIndex(lOwnerSub, vOwners)
        end if
        for j = 1 to length(vSets) do
            lOwnerSub = find(sets[i], vSets[j])
            if lOwnerSub != 0 then
                vSets[j] = removeIndex(lOwnerSub,
                                         vSets[j])
                exit
            end if
        end for
    end for

end procedure

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/proc release_all_mem()
--/desc Free all the acquired memory
-- This gives back to the system, all the memory acquired by calling
-- /acquire_mem(). You must not use any previously acquired memory blocks
-- after this has been called. /n
--
-- /b"NOTE:" When using the Win32Lib library, it is /b not required to call
-- this function as that is done automatically when WinMain() completes.
--
-- /b"WARNING:" Calling this before WinMain() has ended will probably cause the
-- the /i Win32Lib routines to crash.
--
-- Example:
--
--/code
--      -- Return all the memory areas
--      release_all_mem()
--
--/endcode

global procedure release_all_mem()
-- No longer needed.
end procedure



-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func allot( object FldDefn )
--/desc Allocate space in structure for a new field.
--/ret SEQUENCE: Definition to allotted memory.
-- /i FldDefn is either a number of bytes to allocate, one of the predefined
-- datatypes (listed below), or a 2-element sequence containing a repeat count
-- and a datatype or length.
--
-- If a number of bytes is supplied, the field is aligned to the next 32-bit
-- boundry before allocation.
--
-- The returned allotment definition is used by /store and /fetch. It has the
-- following structure. /n
-- The definition has four items: /n
-- An /i offset, a /i datatype, a /i"repeat length", and a /i "unit bytesize" /n
-- Allowable types are: /n
-- /li /b Byte: 8 bit value
-- /li /b Int8: Signed 8 bit integer, same as /b Byte.
-- /li /b Word: 16 bit value
-- /li /b Integer: 16 bit value, same as /b Word
-- /li /b Int16: 16 bit value, same as /b Word
-- /li /b Long: Signed 32 bit value
-- /li /b DWord: 32 bit value, same as /b Long
-- /li /b Int32: 32 bit value, same as /b Long
-- /li /b UInt: Unsigned 32 bit value.
-- /li /b Ptr: 32 bit value, same as /b UInt
-- /li /b ULong: 32 bit value, same as /b UInt
-- /li /b Hndl: 32 bit value, a pointer to a pointer
-- /li /b HndlAddr: 32 bit value,
-- /li /b Lpsz: Long pointer (32 bits) to zero delimited string
-- /li /b Strz: Fixed size buffer that holds a zero-delim string
-- /li /b Single: 32-bit IEEE floating point value
-- /li /b Float: Same as Single
-- /li /b Double: 64-bit IEEE floating point value
-- 
-- Example:
--
-- /code
-- constant
--    msLeft           = allot( Long ),
--    msTop            = allot( Long ),
--    msRight          = allot( Long ),
--    msBottom         = allot( Long ),
--    msVelocity       = allot( Single ),
--    msXYZ            = allot( {4, DWord} ),
--    msReserved       = allot( 5 ),
--    msName           = allot( Lpsz ),
--    msBuffer         = allot( {128, Strz} ),
--    SIZEOF_MYSTRUCT  = /allotted_size()
-- /endcode

global function allot( object pDataType )

    integer soFar, diff, size, i, lCnt

    if sequence(pDataType) then
        lCnt = pDataType[1]
        i =   pDataType[2]
    else
        lCnt = 1
        i   = pDataType
    end if
    
    -- save position    
    soFar = vAllotted

    -- if not a pre-defined type, make sure it gets a word boundary
    size = find(i, vSizeNames)
    if size = 0
    then
       if i > 0 then
          diff = remainder( soFar, 4 )
          if diff then
            -- word align
            soFar = soFar + 4 - diff
          end if

        -- size is actual size
        size = i
       end if
    else
        size = vSizeLengs[size]
    end if

    -- allot space
    vAllotted += (size * lCnt)

    
    -- return offset, data type, count, and unit-size
    return { soFar, i, lCnt, size }
    
end function

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func allotted_handle(hDefn)
--/desc Returns the address of the supplied handle, but with fetch instructions
-- to get the address from the handle, rather than the handle itself. /n
-- An empty sequence is returned if the parameter was invalid.
--/ret SEQUENCE: Handle's "offset" into a structure.
--
-- Example:
--
-- /code                                   
-- constant
--    hDemo            = /allot( Hndl ),
--    pDemo            = /allotted_handle( hDemo ),
--    SIZEOF_DEMO      = /allotted_size()
--     . . .
--    x = allocate_struct(SIZEOF_DEMO)
--    initDEMO(x)
--
--    h = fetch(x, hDemo)
--    a = fetch(x, pDemo)
--    -- 'h' will contain the handle, and 'a' the address from the handle.
-- /endcode


global function allotted_handle(sequence pHandle)
   if length(pHandle) = 4
         and
      pHandle[2] = Hndl
   then
      return {pHandle[1], HndlAddr, pHandle[3], pHandle[4]}
   else
      return {}
   end if
         
    
end function

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func allotted_size()
--/desc Return allocate size of structure, and reset for new structure.
--/ret INTEGER: Allotted size of structure.
--
-- Example:
--
-- /code                                   
-- constant
--    rectLeft            = /allot( Long ),
--    rectTop             = /allot( Long ),
--    rectRight           = /allot( Long ),
--    rectBottom          = /allot( Long ),
--    SIZEOF_RECT        = /allotted_size()
-- /endcode

global function allotted_size()

    -- returns allotted size, and clears size

    integer soFar
    
    soFar = vAllotted

    vAllotted = 0
    
    return soFar
    
end function
    

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/proc store( structure, field, value )
--/desc Store a value into a structure.
-- Type conversion is automatic. For example, if an /b Lpsz field is 
-- used, the value is automatically converted from a sequence to a 
-- C-style string, and the address of that string is stored in the 
-- structure.
--
-- Example:
--
--/code
--      -- allocate RECT structure, and populate it
--      atom rect                                  
--
--      -- allocate the structure
--      rect = /allocate_struct( SIZEOF_RECT )
--
--      -- store values into the structure
--      /store( rect, rectLeft,   x1 )
--      /store( rect, rectTop,    y1 )
--      /store( rect, rectRight,  x2 )
--      /store( rect, rectBottom, y2 )
--
--      -- Here we store individual elements to an /allot array.
--      constant bCoords = allot( {4, Long} )
--      . . . 
--      /store( rect, bCoords & 1, Col)
--      /store( rect, bCoords & 2, Row)
--/endcode

global procedure store( atom struct, sequence s, object o )

    -- Store the data based on its type
    integer where, datatype, lCnt
    atom    at
    sequence bytes
    
    -- rest data type
    datatype = s[2]
    lCnt     = s[3]
    if length(s) > 4 and integer(s[5]) and s[5] > 0 and s[5] <= lCnt then
        s[5] = (s[5]-1) * s[4]
    else
        s   &= 0
    end if    
    where    = s[1] + struct + s[5]

    -- For sequences, make sure no more than lCnt elements are stored.
    if sequence(o) then
        if length(o) < lCnt then
            lCnt = length(o)
        end if
    end if
    
    -- read, based on datatype
    if      datatype = Byte then
        -- poke a byte
        if atom(o) then
            SafePoke( where, o )
        else
            SafePoke(where, o[1 .. lCnt])
        end if
                
    elsif   datatype = Word then
        -- poke a word
            bytes = int_to_bytes( o )
            SafePoke( where, bytes[1 .. 2] )
                
    elsif   datatype = Long then
        -- poke long(s)
        if atom(o) then
            SafePoke4( where, o )
        else
            SafePoke4(where, o[1 .. lCnt])
        end if

    elsif   datatype = UInt then
        -- poke long(s)
        if atom(o) then
            SafePoke4( where, o )
        else
            SafePoke4(where, o[1 .. lCnt])
        end if

    elsif   datatype = Lpsz then
        -- if atom, treat as long        
        if atom( o ) then
            SafePoke4( where, o )
        else
            -- poke the address of the allotted string
            SafePoke4( where, acquire_mem( struct, o  ) )
        end if
        
    elsif   datatype = Hndl then
        -- poke handle(s)
        if atom(o) then
            SafePoke4( where, o )
        else
            SafePoke4(where, o[1 .. lCnt])
        end if

    elsif   datatype = HndlAddr then
        -- poke a handle's address
        at = acquire_mem(struct, 4)
        SafePoke4( at, o)
        SafePoke4( where, at )

    elsif  datatype = Strz then
        -- poke a zero-terminated string
        bytes = o & 0
        if lCnt < s[3] then
            lCnt += 1
        end if
        SafePoke(where, bytes[1 .. lCnt])

    elsif  datatype = Single then
        if atom(o) then
            o = {o}
        end if
        if lCnt > length(o) then
            lCnt = length(o)
        end if
        for i = 1 to lCnt do
            SafePoke(where, atom_to_float32(o[i]))
            where += 4
        end for
                
    elsif  datatype = Double then
        if atom(o) then
            o = {o}
        end if
        if lCnt > length(o) then
            lCnt = length(o)
        end if
        for i = 1 to length(o) do
            SafePoke(where, atom_to_float64(o[i]))
            where += 8
        end for
                
    else
        -- poke a string
        SafePoke( where, o[1 .. lCnt] )
    end if

end procedure



-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func zStringLength( address )
--/desc Get the length of a C-style string.
--/ret INTEGER: The length of the string pointed to by /i address
--
-- Example:
--
--/code
--      -- get length of a C-string from address
--      integer len
--
--      len = zStringLength( address )
--/endcode
global function zStringLength(atom a)

    integer l

    -- Only deal with non-zero addresses
    if a then
        l = c_func(xlstrlen,{a})
    else
        l = 0
    end if

    -- send back the size
    return l
end function

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func peek_string( address )
--/desc Get sequence from address holding C-style string.
--/ret SEQUENCE: containing the C-style string.
-- This is typically done automatically by the /fetch
-- function.
--
-- Example:
--
--/code
--      -- get a C-string from address
--      sequence s
--
--      s = /peek_string( address )
--/endcode
global function peek_string(atom a)
-- V0.56 Al Getz
    integer l
    sequence s

    -- Only deal with non-zero addresses
    if a then
        l = c_func(xlstrlen,{a})
        s = SafePeek({a, l})
    else
        s = {}
    end if

    -- send back all the bytes found.
    return s

end function

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func peek_handle( handle )
--/desc Get address from a Handle.
--/ret ATOM: address.
-- This is typically done automatically by the /fetch
-- function.
--
-- Example:
--
--/code
--      -- get the address from handle
--      atom a, h
--
--      a = /peek_handle( h )
--/endcode


global function peek_handle(atom a)
    return SafePeek4u(a)
end function

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func fetch( structure address, field )
--/desc Fetch field from structure.
--/ret OBJECT: Field from a structure.
-- Data conversion is automatic. For example, if the field is 
-- an /b Lpsz, a sequence containing the string will automatically 
-- be returned.
--
-- Example:
-- /code                                  
--      -- fetch the average character width from the text metrics structure
--      width = /fetch( tm, tmAveCharWidth )
--
--      -- Here we get individual elements from an /allot array.
--      constant bCoords = allot( {4, Long} )
--      . . . 
--      x = /fetch( rect, bCoords & 1)
--      y = /fetch( rect, bCoords & 2)
--
--      -- To get all the values at once
--      sequence coords
--      coords = /fetch( rect, bCoords)
--
-- /endcode

global function fetch( atom struct, sequence s )

    -- fetch the data based on the type
    integer size, char, cnt
    atom at
    object result
    sequence sbytes

    -- get data type
    size    = s[2]
    -- get repeation
    cnt     = s[3]
    if length(s) > 4 and integer(s[5]) and s[5] > 0 and s[5] <= cnt then
        s[5] = (s[5]-1) * s[4]  
        cnt  = 1 -- Force a fetch of a single element.
    else
        s   &= 0
    end if    
    -- address is struct + offset
    at = s[1] + struct + s[5]
    
    -- read, based on size
    if      size = Byte then    
        -- return byte
        if cnt > 1 then
            return (SafePeek( {at, cnt}) )
        else
            return SafePeek( at )
        end if

    elsif   size = Word then
        if cnt = 1 then
            -- return word
            return bytes_to_int( SafePeek({at, 2}) & {0, 0} )
        else
            s = {}
            for i = 1 to cnt do
                s &= bytes_to_int( SafePeek({at, 2}) & {0, 0} )
                at += 2
            end for
            return s
        end if
        
    elsif   size = Long then
        if cnt = 1 then
            -- return signed long
            return SafePeek4s( at )
        else
            s = SafePeek4s({at, cnt} )
            return s
        end if

    elsif   size = UInt then
        if cnt = 1 then
            -- return unsigned long
            return SafePeek4u( at )
        else
            s = SafePeek4u({at, cnt} )
            return s
        end if

    elsif   size = Lpsz then
        -- get the pointer
        at = SafePeek4u(at)
        if not c_func(xIsBadReadPtr,{at,4})then
            -- return the string
            return peek_string( at )
        else
            return at
        end if

    elsif   size = Hndl then
        if cnt = 1 then
            -- return handle
            return SafePeek4u( at )
        else
            return SafePeek4u({at,cnt})
        end if

    elsif   size = HndlAddr then
        -- return a handle's address
        at = SafePeek4u( at )
        at = peek_handle(at)
        return at
    elsif size = Strz then

        return peek_string(at)
                
    elsif size = Single then
        result = {}
        for i = 1 to cnt do
            sbytes = SafePeek({at, 4} )
            at += 4
            result &= float32_to_atom(sbytes)
        end for
        if cnt = 1 then
            result = result[1]
        end if
        return result

    elsif size = Double then
        result = {}
        for i = 1 to cnt do
            sbytes = SafePeek({at, 8} )
            at += 8
            result &= float64_to_atom(sbytes)
        end for
        if cnt = 1 then
            result = result[1]
        end if
        return result

    elsif size > 0 then
        -- return the string                        
        return( SafePeek( {at, size} ) )
    else
        return {}
    end if

end function

-----------------------------------------------------------------------------
--/topic Memory Management Routines
--/func address( structure address, field )
--/desc Get address of /i field in structure.
--/ret ATOM: Address of the field in the structure.
-- This is typically used if the structure contains an array.
--
-- In this snippet, the /b memBitmapInfo structure contains an array
-- of /b RGBQUAD colors. The array is populated with the values in 
-- the pal:
--
-- /code 
--    -- get the start of the rgbQuad array
--    rgbQuad = address( memBitmapInfo, bmiColors )
--
--    -- copy the pal to memory    
--    for i = 1 to colors do
--
--        -- store values
--        /store( rgbQuad, rgbRed,      pal[i][1] )
--        /store( rgbQuad, rgbGreen,    pal[i][2] )
--        /store( rgbQuad, rgbBlue,     pal[i][3] )
--        /store( rgbQuad, rgbReserved, 0 )
--        
--        -- move to next quad
--        rgbQuad += SIZEOF_RGBQUAD
--
--    end for
--
-- /endcode      

global function address( atom addr, object offset )
    -- return address in structure  
    if atom( offset ) then
        return addr + offset
    elsif length(offset) >= 3 then
        if length(offset) > 4 and integer(offset[5]) and offset[5] > 0 and offset[5] <= offset[3] then
            offset[5] = (offset[5]-1) * offset[4]  
        else
            offset   &= 0
        end if    
        -- address is struct + offset
        return offset[1] + addr + offset[5]
    else
        return 0
    end if
end function


--/topic Memory Management Routines
--/func new_memset( a )
--/desc Allocates a unique id for a memory set.
--/ret ATOM: An id for a new memory set (memset).
-- A memset id is actually a machine address of a 4-bytes location. You can
-- use this 4-byte area for anything you like, until you call /release_mem()
--
-- Example:
--
--/code
--      atom ss
--
--		ss = new_memset()
--      b = acquire_mem(ss, "All you need is love")
--		...
--      release_mem( ss )  -- Let go of set 'ss'
--/endcode
global function new_memset()
atom ms
	ms = acquire_mem( 0, UInt )
    return ms
end function

